/*
 *  ======== graph_stacks.js ========
 *  Graph example template for an ROV2 addon element
 *
 *  See also:
 *      https://www.polymer-project.org/1.0/docs/devguide/feature-overview
 *  and for Google Charts:
 *      https://elements.polymer-project.org/elements/google-chart
 */

/* Polymer element registration */
Polymer({
    /* identify this code's element: must match <dom-module> id */
    is: "xdc-rov-polymerUI-examples-graph_stacks",

    /* common view behaviors. See ti-rov-behaviors.html for reference */
    behaviors: [rovBehaviors.viewBehaviors],

    /* properties for this element's public API */
    properties: {

        graph: {       /* the google bar graph element */
          type:  Object,
          value: {}
        },

        rows: {        /* each row = array representing a group of bars */
          type:  Array,
          value: []
        },

        viewName: {    /* required read-only addon view name */
          type:  String,
          value: "Stacks Graph"
        },

        viewRefresh: { /* recommended view state indicator */
          type:  Boolean,
          value: false
        },

        hwiTotal: {       /* total HWI stack space used (bytes) */
            type:  Number,
            value: 0
        },

        taskTotal: {       /* total task stack space used (bytes) */
            type:  Number,
            value: 0
        },

        OVERFLOW_MSG: {
            type:  String,
            value: "Stack overflow likely corrupted adjacent data!"
        }
    },

    /* methods on the element's prototype */

    /*
     *  ======== updateGraph ========
     *  Redraw the graph based on new data or new size
     */
    updateGraph: function (graph, rows) {
        /* remove all rows */
        graph._dataTable.removeRows(0, rows.length);

        /* add all current stack row data */
        for (var i = 0, total = 0; i < rows.length; i++) {
            graph._dataTable.addRow(rows[i]);
        }

        graph.options.hAxis.viewWindow.max = rows.length;

        /* update hAxis title to show total stack usage */
        graph.options.hAxis.title = "Stack (Total used: "
            + (this.taskTotal + this.hwiTotal) + " Bytes)";

        /* redraw graph so changes become visible */
        graph.drawChart();
    },

    /*
     *  ======== getViewCallback ========
     *  Callback from rovData.getView()
     *
     *  Params
     *    error      - null when there is no error; otherwise an error
     *                 message string
     *    viewData   - the ROV view data requested as a JSON object
     *    moduleName - module name string for the returned view data
     *    viewName   - view name string for the returned view data
     */
    getViewCallback: function (error, viewData, moduleName, viewName) {
        /* show the refresh button and hide the refresh spinner */
        if (this.viewRefresh) {
            this.viewRefresh = false;
        }

        /* handle data acquisition error (if any) */
        if (error != null) {
            return;
        }

        /* create rows representing the data to display */
        if (viewName == "Detailed") {
            /* handle Task Details viewData */
            var taskArray = this._genTaskArray(viewData);
            var total = 0;
            this.rows.splice(1);              /* remove task stack rows */
            for (var i = 0; i < taskArray.length; i++) {
                this.rows.push(taskArray[i]); /* append new tasks stack rows */
            }
        }
        else if (viewName == "Module") {
            /* handle Task Module viewData */
            var hwiRow = this._genHwiRow(viewData);
            this.rows.shift();                /* remove old Hwi stack row */
            this.rows.unshift(hwiRow);        /* prepend new Hwi stack row */
        }

        /* update graph with latest data */
        this.updateGraph(this.$.graph, this.rows);
    },

    /*
     *  ======== _chartReady ========
     *  This listener is fired when chart is ready for interaction.
     *
     *  We use this for the initial update when the graph is first
     *  opened. The google-charts docs say this listener is a 'one time'
     *  event, but it fires endlessly, so instead of simply assigning it
     *  in the html, we set it programatically in ready() and remove it
     *  here after one call.
     */
    _chartReady: function (chart) {
        /* remove listener used for intitial refresh */
        this.unlisten(this.$.graph, 'google-chart-render', '_chartReady');

        /* trigger our refresh of the data */
        this.onRefresh();
    },

    /*
     *  ======== _genHwiRow ========
     */
    _genHwiRow: function (viewData) {
        var size = viewData.hwiStackSize - 0;
        var peak = viewData.hwiStackPeak - 0;

        var overflow = 0;
        var used, free;
        if (peak >= size) {
            used = free = 0;
            overflow = 1;
        }
        else {
            used = peak / size;
            free = 1 - used;
        }

        /* create the row for display */
        var row = ["ISR Stack", used, peak + " B", free, overflow, "", ""];
        if (overflow == 1) {
            row[2] = "";                 /* remove annotation for % used */
            row[5] = "> " + size + " B"; /* annotate overflow */
            row[6] = this.OVERFLOW_MSG;  /* overflow tooltip */
        }

        /* update latest ISR stack usage */
        this.hwiTotal = peak - 0;

        return (row);
    },

    /*
     *  ======== _genTaskArray ========
     */
    _genTaskArray: function (viewData) {
        var taskArray = [];
        var total = 0;

        /* create rows representing the data to display */
        for (var i = 0; i < viewData.length; i++) {
            var task = viewData[i];
            var size = task.stackSize - 0;
            var peak = task.stackPeak - 0;

            var elem = ["", 0, peak + " B", 0, 0, "", ""];
            elem[0] = task.label == "" ? task.fxn[0] : task.label;
            if (elem[0]) { /* trim long names */
                var k = elem[0].lastIndexOf('.');
                if (k > 10) {
                    elem[0] = elem[0].substr(k + 1);
                }
            }

            if (size == 0) { /* pre-initialized task */
                elem[1] = 1;                  /* used */
                elem[3] = 0;                  /* free */
                elem[4] = 0;                  /* overflow */
            }
            else if (peak >= size) {
                elem[1] = elem[3] = 0;        /* overflow detected! */
                elem[2] = "";                 /* clear used annotation */
                elem[4] = 1;                  /* set overflow to 100% */
                elem[5] = "> " + size + " B"; /* overflow annotation */
                elem[6] = this.OVERFLOW_MSG;  /* overflow tooltip */
            }
            else {
                elem[1] = peak / size;        /* used */
                elem[3] = 1 - elem[1];        /* free */
                elem[4] = 0;                  /* overflow */
            }
            taskArray.push(elem);
            total += peak;
        }

        /* sort task stack results by "stack name" */
        taskArray.sort(function (a, b) {
            if (a[0] < b[0]) {
                return 1;
            }
            else if (a[0] > b[0]) {
                return (-1);
            }
            return (0);
        });

        /* update latest total task stack usage */
        this.taskTotal = total;

        return (taskArray);
    },

    /*
     *  ======== onRefresh ========
     *  Called when user clicks the 'Refresh' button on this view,
     *  or 'Refresh all' or 'Repeat refresh' buttons on the ROV toolbar
     */
    onRefresh: function () {
        /* hide the refresh button and show the refresh spinner */
        this.viewRefresh = true;

        /*
         *  Request Task Details ROV view.
         *
         *  This is an asynchronous call; getViewCallback() will be called
         *  with the result in its viewData parameter.
         */
        this.rovData.getView("ti.sysbios.knl.Task", "Detailed", this);

        /* also request Task Module to get HWI stack usage */
        this.rovData.getView("ti.sysbios.knl.Task", "Module", this);
    },

   /*
    *  ======== onSave ========
    *  Called when user clicks the 'Download' button on this view,
    *  or the 'Download all' button on the ROV toolbar
    */
    onSave: function () {
        /* create an object representing the graph's current data set */
        this.saveView(JSON.stringify(this.rows, null, '\t'));
    },

    /*
     *  ======== ready ========
     *  Called by Polymer when view is created but only after property values
     *  are set and the local DOM is initialized.
     *
     *  ti-rov-panel is the global container element for ROV.  It has
     *  already retrieved the modules and view list, so we can get these
     *  directly from there instead of making a redundant asynchronous
     *  calls into ti-rov-data.
     *
     *  To get the ti-rov-panel element:
     *        var rovPanel = document.querySelector('ti-rov-panel');
     */
    ready: function () {
        /*
         *  Configure the google-chart element. For reference, see:
         *      https://elements.polymer-project.org/elements/google-chart
         */
        var graph = this.$.graph;
        graph.options.isStacked = 'percent';
        graph.options.vAxis = {minValue: 0, ticks: [0, .25, .50, .75, .9, 1]};

        graph.options.hAxis = {title: "Stack", viewWindow: {min: 0, max: 1}};

        /* define chart area */
        graph.options.chartArea = {top: 30, bottom: 40, left: 50, right: 40};

        /* enable horizontal scrolling via mouse drag */
        graph.options.explorer = {axis: 'horizontal'};

        /* place the legend at top, on a single line */
        graph.options.legend = {position: 'top', maxLines: 1};

        /* place the legend at top, on a single line */
        graph.options.bar = {groupWidth: "75%"};

        /* define colors for each bar catagory */
        graph.options.series = [
            {color: 'blue',    visibleInLegend: true}, /* used color */
            {color: '#e6ac00', visibleInLegend: true}, /* free color */
            {color: 'red',     visibleInLegend: true}  /* overflow color */
        ];

        /* define bar graph's data row types */
        graph.cols = [
            {label: "Stack",      type: "string"}, /* stack "name" */
            {label: "% Used",     type: "number"}, /* used % value */
            {role:  "annotation", type: "string"}, /* show used value */
            {label: "% Free",     type: "number"}, /* free % value */
            {label: "Overflow",   type: "number"}, /* overflow % value */
            {role:  "annotation", type: "string"}, /* show overflow value */
            {role:  "tooltip",    type: "string"}  /* overflow tooltip */
        ];

        graph.rows = [ ["", 0, "", 0, 0, "", ""] ];

        /* set the graph to its min size */
        graph.style.width = '390px';
        graph.style.height = '264px';

        /* initialize all our state */
        this.graph = graph;
        this.rows = graph.rows;

        /* set listener to set initial data _after_ the chart is ready */
        this.listen(this.$.graph, 'google-chart-render', '_chartReady');
    },

    /*
     *  ======== resized ========
     *  Callback initially passed to rovUtils.initRovResize() from
     *  resizerClicked(). Called after the view has been resized,
     *  enabling us to resize the graph accordingly
     */
    resized: function () {
        /* get viewPaperCard defined via the element in graph_stacks.html */
        var viewPaperCard = this.$.viewPaperCard;
        var graph = this.$.graph;

        /* get the new width and height from viewPaperCard's style */
        var width = Number(viewPaperCard.style.width.slice(0, -2));
        var height = Number(viewPaperCard.style.height.slice(0, -2));

        /* get viewPaperCard minWidth and minHeight */
        var minWidth = Number(viewPaperCard.style.minWidth.slice(0, -2));
        var minHeight = Number(viewPaperCard.style.minHeight.slice(0, -2));

        /* Don't let new size go below the defined minimums */
        width = Math.max(width, minWidth);
        height = Math.max(height, minHeight);

        /* Scale the graph accordingly */
        graph.style.width = (width * .95) + 'px';
        graph.style.height = (height * .85) + 'px';

        /* don't let the width get too small */
        if (width < minWidth) {
            width = minWidth;
        }

        /* update the graph and redisplay it */
        this.updateGraph(graph, this.rows);
    },

    /*
     *  ======== resizerClicked ========
     *  Called on a bottom corner on-mousedown event
     *
     *  See graph_stacks.html resizer div.
     */
    resizerClicked: function (e) {
        /*
         *  call polymerUI/src/rovUtils.js initRovResize(), passing the
         *  viewPaperCard element that is being resized. Also pass callback
         *  object in order to be notified via this.resized() when the element
         *  is done being is resized, so we can make any required size related
         *  changes to the graph itself
         */
        rovUtils.initRovResize(e, this.$.viewPaperCard, {elem: this, fxn: this.resized});
    }
});
